/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "moab/ScdInterface.hpp"
#include "MBMesh.hpp"
#include "PNclMBUtils.hpp"
#include "MeshTypes.hpp"

moab::DataType PNclNCMapToMB(const nc_type type){
	moab::DataType dt = moab::MB_MAX_DATA_TYPE;
	switch(type){
		case NC_DOUBLE:
			dt = moab::MB_TYPE_DOUBLE;
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Cannot convert nc_type to a supported maob data type");
	}
	return dt;
}

nc_type PNclMBMapToNC(const moab::DataType mb_type)
{
	nc_type dt = NC_EBADTYPE;
	switch(mb_type){
		case moab::MB_TYPE_DOUBLE:
			dt = NC_DOUBLE;
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Cannot convert moab data type to nc_type");
	}
	return dt;
}

NclBasicDataTypes PNclMBMapToNcl(const moab::DataType mb_type)
{
	NclBasicDataTypes dt = NCL_none;
	switch(mb_type){
		case moab::MB_TYPE_DOUBLE:
			dt = NCL_double;
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Cannot convert moab data type to NCL type");
	}
	return dt;
}

/* FIXME: Create tag is currently unusable - need to iterate MB tag to explicitly allocate storage */
/* FIXME: Add dimension as a parameter to the func  and change "verts" to "ents" */
moab::ErrorCode PNclMBCreateTag(parncl::MBMesh *mesh, const char *tag_name, size_t tstep, size_t nlevs, void *init_val, moab::DataType dt, moab::EntityType ent_type)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::Tag mb_tag;
	moab::Range rents;
	void *from_data=NULL;
	std::ostringstream tag_name_ostr;
	int sz = 0;

	moab::Interface *mb_core = mesh->get_mb_core();
	moab::EntityHandle mb_set = mesh->get_mb_mset();

	tag_name_ostr << tag_name << tstep;

	mb_retval = mb_core->get_entities_by_type(mb_set, ent_type, rents);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB 0-dim entities for creating (%s), err=%d", tag_name, mb_retval);
		return mb_retval;
	}
	sz = rents.size();

	/* FIXME: Do we create all tags with the same 0-dim size */
	bool mb_tag_created;
	mb_retval = mb_core->tag_get_handle(tag_name_ostr.str().c_str(), nlevs, dt, mb_tag, moab::MB_TAG_CREAT|moab::MB_TAG_DENSE, init_val, &mb_tag_created);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error creating moab tag for (%s), err=%d", tag_name, mb_retval);
		return mb_retval;
	}
	assert(mb_tag_created);

	/* Explicitly allocate memory for the tag by iterating over it */
	mb_retval = mb_core->tag_iterate(mb_tag, rents.begin(), rents.end(), sz, from_data);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB iterators for (%s), err=%d", tag_name, mb_retval);
		return mb_retval;
	}

	return mb_retval;
}

moab::ErrorCode PNclMBCreateSparseTag(parncl::MBMesh *mesh, const char *tag_name, size_t tag_sz, void *init_val, moab::DataType dt, moab::EntityType ent_type)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::Tag mb_tag;
	void *tag_data=NULL;
	std::ostringstream tag_name_ostr;
	int sz = static_cast<int> (tag_sz);

	moab::Interface *mb_core = mesh->get_mb_core();
	moab::EntityHandle mb_set = mesh->get_mb_mset();

	if(sz <= 0){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Invalid size (0) specified for var dims");
		return moab::MB_FAILURE;
	}


	/* FIXME: Do we create all tags with the same 0-dim size */
	bool mb_tag_created;
	//mb_retval = mb_core->tag_get_handle(tag_name, sz, dt, mb_tag, moab::MB_TAG_CREAT | moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN, init_val, &mb_tag_created);
	mb_retval = mb_core->tag_get_handle(tag_name, sz, dt, mb_tag, moab::MB_TAG_CREAT | moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error creating moab tag for (%s), err=%d", tag_name, mb_retval);
		return mb_retval;
	}
	//assert(mb_tag_created);

	switch(dt){
		case moab::MB_TYPE_DOUBLE:
				tag_data = NclCalloc(sz, sizeof(double));
				if(tag_data == NULL){
					NhlPError(NhlFATAL, NhlEUNKNOWN, "Could not alloc mem for tag");
					return moab::MB_FAILURE;
				}
				break;
		default:
				NhlPError(NhlFATAL, NhlEUNKNOWN, "Invalid type specified for tag");
				return moab::MB_FAILURE;
	}

	mb_retval = mb_core->tag_set_by_ptr(mb_tag, &mb_set, 1, &tag_data, &sz);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error setting data for moab tag for (%s), err=%d", tag_name, mb_retval);
		return mb_retval;
	}

	return mb_retval;
}

moab::ErrorCode PNclMBLoadSparseTags(parncl::MBMesh *mesh, size_t fidx)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	assert(mesh != NULL);
	moab::Interface *mb_core = mesh->get_mb_core();
	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();
	moab::EntityHandle &file_set = pcal_finfo.get_file_set();

	std::string mesh_fload_opts = mesh_finfo.get_mb_sload_opts();

	const std::vector<std::string> &nc_fnames = mesh_finfo.get_nc_fnames();

	mb_retval = mb_core->load_file(nc_fnames[fidx].c_str(), &file_set, mesh_fload_opts.c_str());
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error loading file (%s) in MOAB, err=%d(%s)", nc_fnames[fidx].c_str(), mb_retval, mb_core->get_error_string(mb_retval).c_str());
		return mb_retval;
	}

	return mb_retval;
}

moab::ErrorCode PNclMBRdSparseTag(parncl::MBMesh *mesh, const char *tag_name, moab::DataType dt, void *&val_buf, int &val_len)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::Tag val_tag;
	const void *val_data;

	assert(mesh != NULL);
	moab::Interface *mb_core = mesh->get_mb_core();

	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle &file_set = pcal_finfo.get_file_set();

	mb_retval = mb_core->tag_get_handle(tag_name, val_len, dt, val_tag, moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN);
	if(mb_retval != moab::MB_SUCCESS){
		// Retry after hard loading
		std::string mesh_fload_opts = mesh_finfo.get_mb_sload_opts();
		mesh_fload_opts += std::string(tag_name);

		const std::vector<std::string> &nc_fnames = mesh_finfo.get_nc_fnames();

		// FIXME: This is wrong... We are only loading the first file here...
		// Let the upper layer decide on loading the files
		mb_retval = mb_core->load_file(nc_fnames[0].c_str(), &file_set, mesh_fload_opts.c_str());
		if(mb_retval == moab::MB_SUCCESS){
			mb_retval = mb_core->tag_get_handle(tag_name, val_len, dt, val_tag, moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN);
		}
	}
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB tag for (%s), err=%d(%s)", tag_name, mb_retval, mb_core->get_error_string(mb_retval).c_str());
		return mb_retval;
	}

	mb_retval = mb_core->tag_get_by_ptr(val_tag, &file_set, 1, &val_data, &val_len);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error getting tag buf for (%s), err=%d(%s)", tag_name, mb_retval, mb_core->get_error_string(mb_retval).c_str());
		return mb_retval;
	}

	switch(dt){
		case moab::MB_TYPE_DOUBLE:
			memcpy(val_buf, val_data, val_len * sizeof(double));
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unsupported type, unable to copy data");
			return moab::MB_FAILURE;
	}

	return mb_retval;
}

moab::ErrorCode PNclMBGetSparseTag(parncl::MBMesh *mesh, const char *tag_name, moab::DataType dt, void *&val_buf_ptr, int &val_len)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::Tag val_tag;
	const void *val_data;

	assert(mesh != NULL);
	moab::Interface *mb_core = mesh->get_mb_core();

	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle &file_set = pcal_finfo.get_file_set();

	//mb_retval = mb_core->tag_get_handle(tag_name, val_len, dt, val_tag, moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN);
	mb_retval = mb_core->tag_get_handle(tag_name, val_len, moab::MB_TYPE_OPAQUE, val_tag, moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN);
	if(mb_retval != moab::MB_SUCCESS){
		// Retry after hard loading
		std::string mesh_fload_opts = mesh_finfo.get_mb_sload_opts();
		mesh_fload_opts += std::string(tag_name);

		const std::vector<std::string> &nc_fnames = mesh_finfo.get_nc_fnames();

		// FIXME: This is wrong... We are only loading the first file here...
		// Let the upper layer decide on loading the files
		mb_retval = mb_core->load_file(nc_fnames[0].c_str(), &file_set, mesh_fload_opts.c_str());
		if(mb_retval == moab::MB_SUCCESS){
			mb_retval = mb_core->tag_get_handle(tag_name, val_len, moab::MB_TYPE_OPAQUE, val_tag, moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN);
		}
		else{
			printf("load file failed (%s)", mb_core->get_error_string(mb_retval).c_str());
		}
	}
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB tag for (%s), err=%d(%s)", tag_name, mb_retval, mb_core->get_error_string(mb_retval).c_str());
		return mb_retval;
	}

	mb_retval = mb_core->tag_get_by_ptr(val_tag, &file_set, 1, &val_data, &val_len);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error getting tag buf for (%s), err=%d(%s)", tag_name, mb_retval, mb_core->get_error_string(mb_retval).c_str());
		return mb_retval;
	}

	switch(dt){
		case moab::MB_TYPE_DOUBLE:
			/* FIXME: This is ugly */
			val_buf_ptr = const_cast<void *> (val_data);
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unsupported type, unable to copy data");
			return moab::MB_FAILURE;
	}

	return mb_retval;
}

/* The function assumes that data corresponding to from_tag is already loaded */
moab::ErrorCode PNclMBDupTag(parncl::MBMesh *mesh, const char *to_tag_name, size_t to_tstep, const char *from_tag_name, size_t from_tstep, size_t nlevs, moab::DataType dt, moab::EntityType ent_type)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::Tag to_tag, from_tag;
	void *from_data, *to_data=NULL;
	int len = 0;

	assert(mesh != NULL);
	moab::Interface *mb_core = mesh->get_mb_core();
	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::Range rents;
	moab::EntityHandle &file_set = pcal_finfo.get_file_set();

	mb_retval = mb_core->get_entities_by_type(file_set, ent_type, rents);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB entities, err=%d", mb_retval);
		return mb_retval;
	}

	// Get from from tag characteristics

	std::ostringstream from_tag_name_ostr, to_tag_name_ostr;
	from_tag_name_ostr << from_tag_name << from_tstep;
	to_tag_name_ostr << to_tag_name << to_tstep;
	len = rents.size();

	mb_retval = mb_core->tag_get_handle(from_tag_name_ostr.str().c_str(), nlevs, dt, from_tag, moab::MB_TAG_DENSE);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB tag for (%s), err=%d", from_tag_name, mb_retval);
		return mb_retval;
	}

	mb_retval = mb_core->tag_iterate(from_tag, rents.begin(), rents.end(), len, from_data);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB iterators for (%s), err=%d", from_tag_name, mb_retval);
		return mb_retval;
	}

	// Create a new to tag and copy data
	mb_retval = mb_core->tag_get_handle(to_tag_name_ostr.str().c_str(), nlevs, dt, to_tag, moab::MB_TAG_CREAT|moab::MB_TAG_DENSE);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error creating moab tag for (%s), err=%d", to_tag_name, mb_retval);
		return mb_retval;
	}

	mb_retval = mb_core->tag_iterate(to_tag, rents.begin(), rents.end(), len, to_data);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB iterators for (%s), err=%d", from_tag_name, mb_retval);
		return mb_retval;
	}

	switch(dt){
		case moab::MB_TYPE_DOUBLE:
			memcpy(to_data, from_data, nlevs * len * sizeof(double));
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unsupported type, unable to copy data");
			return moab::MB_FAILURE;
	}

	return mb_retval;
} 

moab::ErrorCode PNclMBDupSparseTag(parncl::MBMesh *mesh, const char *to_tag_name, const char *from_tag_name, moab::DataType dt, moab::EntityType ent_type)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::Tag to_tag, from_tag;
	void *from_data, *to_data=NULL;
	int len = 0;

	assert(mesh != NULL);
	moab::Interface *mb_core = mesh->get_mb_core();
	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle &file_set = pcal_finfo.get_file_set();

	mb_retval = PNclMBGetSparseTag(mesh, from_tag_name, dt, from_data, len);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get sparse tag (%s)", from_tag_name);
		return mb_retval;
	}

	mb_retval = PNclMBCreateSparseTag(mesh, to_tag_name, len, NULL, dt, ent_type);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to create sparse tag (%s)", to_tag_name);
		return mb_retval;
	}

	mb_retval = PNclMBGetSparseTag(mesh, to_tag_name, dt, to_data, len);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get sparse tag (%s)", to_tag_name);
		return mb_retval;
	}

	switch(dt){
		case moab::MB_TYPE_DOUBLE:
			memcpy(to_data, from_data, len * sizeof(double));
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unsupported type, unable to copy data");
			return moab::MB_FAILURE;
	}

	return mb_retval;
} 


moab::ErrorCode PNclDBGMBPrintTag(parncl::MBMesh *mesh, const char *tag_name, size_t tstep, size_t nlevs, moab::DataType dt)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::Tag val_tag;
	const void *val_data;
	int val_len=0;
	unsigned int tag_hnd_flags = 0;
	std::ostringstream tag_name_ostr;

	assert(mesh != NULL);
	moab::Interface *mb_core = mesh->get_mb_core();

	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle &file_set = pcal_finfo.get_file_set();

	if(tstep >= 0){
		tag_name_ostr << tag_name << tstep;
		tag_hnd_flags = moab::MB_TAG_DENSE;
	}
	else{
		tag_name_ostr << tag_name;
		tag_hnd_flags = moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN;
	}

	mb_retval = mb_core->tag_get_handle(tag_name_ostr.str().c_str(), val_len, dt, val_tag, tag_hnd_flags);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB tag for (%s), err=%d(%s)", tag_name_ostr.str().c_str(), mb_retval, mb_core->get_error_string(mb_retval).c_str());
		return mb_retval;
	}

	int num_levs = static_cast<int >(nlevs);
	mb_retval = mb_core->tag_get_by_ptr(val_tag, &file_set, num_levs, &val_data, &val_len);
	if(mb_retval != moab::MB_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error getting tag buf for (%s), err=%d(%s)", tag_name, mb_retval, mb_core->get_error_string(mb_retval).c_str());
		return mb_retval;
	}

	std::ostringstream print_ostr;
	switch(dt){
		case moab::MB_TYPE_DOUBLE:
			print_ostr << std::endl;
			for(size_t i=0; i<val_len; i++){
				print_ostr << ((double *) val_data)[i] << " , ";	
			}
			print_ostr << std::endl;
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unsupported type, unable to copy data");
			return moab::MB_FAILURE;
	}
	std::cout << print_ostr.str().c_str();

	return mb_retval;
}

size_t PNclMBGetChunkSz(parncl::MBMesh *mesh, const char *tag_name, moab::DataType dt, moab::EntityType ent_type)
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;

	moab::Interface *mb_core = mesh->get_mb_core();
	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	moab::EntityHandle &mb_set = pcal_finfo.get_file_set();

	if(ent_type != moab::MBENTITYSET){
		moab::Range rents;
		mb_retval = mb_core->get_entities_by_type(mb_set, ent_type, rents);
		if(mb_retval != moab::MB_SUCCESS){
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Error retrieving MOAB chunk size, err=%d", mb_retval);
			return -1;
		}
		return(rents.size());
	}
	else{
		void *val_buf;
		int val_len;
		mb_retval = PNclMBGetSparseTag(mesh, tag_name, dt, val_buf, val_len);
		if(mb_retval != moab::MB_SUCCESS){
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to retreive size of %s (%s)", tag_name, mb_core->get_error_string(mb_retval).c_str());
			return -1;
		}
		return val_len;
	}

}

const std::string CAM_FV_Eul_MOAB_FLOAD_OPTIONS("PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS;PARALLEL_GHOSTS=2.0.1");
const std::string CAM_HOMME_MOAB_FLOAD_OPTIONS("PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS");
const std::string MOAB_DEFAULT_PART_METHOD("ALLJORKORI");
parncl::MBMesh *create_mb_mesh(moab::Interface *mb_gcore, std::vector<std::string> &fnames, MPI_Comm mb_comm, moab::ParallelComm *mb_pcomm, const std::string &part_method = std::string(MOAB_DEFAULT_PART_METHOD))
{
	parncl::MeshFInfo *mesh_sfinfo = new parncl::MeshFInfo(fnames, mb_gcore, mb_pcomm, part_method);
	
	pargal::fileinfo &pcal_finfo = mesh_sfinfo->get_pcal_finfo();	
	const pargal::pcatt& att = pcal_finfo.get_att();
	std::set<std::string> dim_names = pcal_finfo.dim_names();

	if(parncl::CAM_FV_Mesh::type_of(att, dim_names)){
		delete mesh_sfinfo;
		parncl::MeshFInfo *mesh_finfo = new parncl::MeshFInfo(fnames, mb_gcore, mb_pcomm, CAM_FV_Eul_MOAB_FLOAD_OPTIONS, part_method);
		parncl::CAM_FV_Mesh *new_mesh = new parncl::CAM_FV_Mesh(mb_gcore, mb_comm, mb_pcomm, mesh_finfo, part_method);
		return new_mesh;
	}
	else if(parncl::CAM_HOMME_Mesh::type_of(att, dim_names)){
		delete mesh_sfinfo;
		parncl::MeshFInfo *mesh_finfo = new parncl::MeshFInfo(fnames, mb_gcore, mb_pcomm, CAM_HOMME_MOAB_FLOAD_OPTIONS, part_method);
		parncl::CAM_HOMME_Mesh *new_mesh = new parncl::CAM_HOMME_Mesh(mb_gcore, mb_comm, mb_pcomm, mesh_finfo, part_method);
		return new_mesh;
	}
	else if(parncl::CAM_Eul_Mesh::type_of(att, dim_names)){
		delete mesh_sfinfo;
		parncl::MeshFInfo *mesh_finfo = new parncl::MeshFInfo(fnames, mb_gcore, mb_pcomm, CAM_FV_Eul_MOAB_FLOAD_OPTIONS, part_method);
		parncl::CAM_Eul_Mesh *new_mesh = new parncl::CAM_Eul_Mesh(mb_gcore, mb_comm, mb_pcomm, mesh_finfo, part_method);
		return new_mesh;
	}
	else{
		delete mesh_sfinfo;
		parncl::MeshFInfo *mesh_finfo = new parncl::MeshFInfo(fnames, mb_gcore, mb_pcomm, CAM_FV_Eul_MOAB_FLOAD_OPTIONS, part_method);
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to determine type of mesh to create");
		parncl::CAM_Eul_Mesh *new_mesh = new parncl::CAM_Eul_Mesh(mb_gcore, mb_comm, mb_pcomm, mesh_finfo, part_method);
		return new_mesh;
	}	
}
