/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "ScdMBMesh.hpp"
#include "PNclMDUtils.hpp"

namespace parncl{

void ScdMBMesh::find_scd_box(void )
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;

	moab::EntityHandle mb_mset = mesh_finfo_->get_mb_mset();

	mb_retval = mb_gcore_->query_interface(scdi_);
	if((mb_retval != moab::MB_SUCCESS) || !scdi_){
		std::string funcname("UNKNOWN");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to query the MOAB scdi interface");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
	}

	std::vector<moab::ScdBox *> scd_boxes;
	mb_retval = scdi_->get_boxes(scd_boxes);
	if(mb_retval != moab::MB_SUCCESS){
		std::string funcname("UNKNOWN");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get MOAB Scd boxes");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
	}

    bool found = false;
    for(std::vector<moab::ScdBox *>::iterator box_iter = scd_boxes.begin(); box_iter != scd_boxes.end(); ++box_iter){
        moab::ScdBox *scd_box = *box_iter;
        moab::EntityHandle scd_box_set = scd_box->box_set();
        if(mb_gcore_->contains_entities(mb_mset, &scd_box_set, 1)){
            found = true;
			scd_box_ = scd_box;
			break;
        }
    }

    if(!found){
		std::string funcname("UNKNOWN");
        NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get box dims for MOAB mesh set");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
    }
}

void ScdMBMesh::init_ijk_info(void )
{
	moab::ErrorCode mb_retval = moab::MB_SUCCESS;
	moab::EntityHandle mb_mset = mesh_finfo_->get_mb_mset();

	moab::HomCoord h_min = scd_box_->box_min();
	moab::HomCoord h_max = scd_box_->box_max();

	/*
	lijk_dims_.resize(3);
	lijk_dims_[0] = std::make_pair(h_min.i(), h_max.i());
	lijk_dims_[1] = std::make_pair(h_min.j(), h_max.j());
	lijk_dims_[2] = std::make_pair(h_min.k(), h_max.k());
	*/

	int x_sz, y_sz, z_sz;
	scd_box_->box_size(x_sz, y_sz, z_sz);

	const double *xc, *yc, *zc;
	mb_retval = scd_box_->get_coordinate_arrays(xc, yc, zc);
	if(mb_retval != moab::MB_SUCCESS){
		std::string funcname("UNKNOWN");
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to get MOAB coord arrays");
		throw pargal::pargal_except(mb_gcore_->get_error_string(mb_retval), __FILE__, __LINE__, funcname);
	}

	lijk_coords_min_max_.resize(3);
	lijk_coords_min_max_[0] = std::make_pair(xc[0], xc[x_sz - 1]);
	lijk_coords_min_max_[1] = std::make_pair(yc[0], yc[y_sz - 1]);
	lijk_coords_min_max_[2] = std::make_pair(zc[0], zc[z_sz - 1]);
}

ScdMBMesh::ScdMBMesh(moab::Interface *mb_gcore, MPI_Comm mb_comm, moab::ParallelComm *mb_pcomm, MeshFInfo *mesh_finfo, const std::string &part_method):MBMesh::MBMesh(mb_gcore, mb_comm, mb_pcomm, mesh_finfo, part_method)
{
	scdi_ = NULL;
	find_scd_box();
	init_ijk_info();
}

ScdMBMesh::~ScdMBMesh()
{
	if(scdi_){
		mb_gcore_->release_interface(scdi_);
	}
	if(scd_box_){
		delete scd_box_;	
	}
}

const std::string PCAL_VAL_LOC_NAMES[] = {std::string("QUAD")};

bool ScdMBMesh::ctest(const std::vector<size_t>& var_dims, std::string &pcal_val_loc)
{
	bool ctest_result = true;
	std::vector<size_t>::const_reverse_iterator var_dims_iter = var_dims.rbegin();
	const std::vector<std::string> pcal_val_loc_names(PCAL_VAL_LOC_NAMES, PCAL_VAL_LOC_NAMES + sizeof(PCAL_VAL_LOC_NAMES)/sizeof(std::string));
	
	for(std::vector<std::string>::const_iterator pcal_val_loc_names_iter = pcal_val_loc_names.begin(); pcal_val_loc_names_iter != pcal_val_loc_names.end(); ++pcal_val_loc_names_iter){
		pcal_val_loc = *pcal_val_loc_names_iter;
		const std::vector<std::pair<size_t, size_t> > gij_conform_dims = this->get_gij_conform_dims(pcal_val_loc);
		std::vector<std::pair<size_t, size_t> >::const_iterator gij_conform_dims_iter = gij_conform_dims.begin();
		while((var_dims_iter != var_dims.rend()) &&
				(gij_conform_dims_iter != gij_conform_dims.end())){
			std::pair<size_t, size_t> dims = *gij_conform_dims_iter;
			if((dims.second - dims.first) != *var_dims_iter){
				ctest_result = false;
				break;
			}
			++var_dims_iter;
			++gij_conform_dims_iter;
		}
		if((gij_conform_dims_iter != gij_conform_dims.end()) && (ctest_result == true)){
			for(; gij_conform_dims_iter != gij_conform_dims.end(); ++gij_conform_dims_iter){
				std::pair<size_t, size_t> dims = *gij_conform_dims_iter;
				if(dims.second != 0){
					ctest_result = false;
					break;
				}
			}	
		}
		if(ctest_result){
			break;
		}
	}

	return ctest_result;
}

bool ScdMBMesh::ctest(ParCALMeshVal &old_val, NclSelectionRecord *sel)
{
	/*
	std::string funcname("UNKNOWN");
	NhlPError(NhlFATAL, NhlEUNKNOWN, "Not implemented");
	throw pargal::pargal_except(std::string("Not implemented"), __FILE__, __LINE__, funcname);
	*/
	return false;
}

void ScdMBMesh::conv_to_local(const NclSelectionRecord *gsel, NclSelectionRecord *lsel, const std::vector<size_t> &gdim_sz, std::vector<size_t> &ldim_sz, const std::string &pcal_val_loc)
{
	std::vector<std::pair<size_t, size_t> > lijk_dims = this->get_lij_conform_dims(pcal_val_loc);
	if(part_method_ == "SQIJ"){
		size_t n_dims = ldim_sz.size();
		size_t lev_idx = (n_dims - 1) - 2;
		assert(n_dims < gsel->n_entries);
		size_t n_lijk_dims = lijk_dims.size();
		PNclSelCopy(lsel, gsel, 1);
		// We have same number of timesteps and levels on each proc
		for(size_t i=n_dims-1, j=0; (i>=lev_idx) && (j<n_lijk_dims); i--, j++){
			NclSelectionTypes sel_type = lsel->selection[i].sel_type;
			ldim_sz[i] = lijk_dims[j].second - lijk_dims[j].first + 1;
			if(sel_type == Ncl_SUB_ALL){
				lsel->selection[i].u.sub.start = lijk_dims[j].first;
				lsel->selection[i].u.sub.finish = lijk_dims[j].second;
				lsel->selection[i].u.sub.stride = 1;	
			}
			else if(sel_type == Ncl_SUBSCR){
				size_t stride = lsel->selection[i].u.sub.stride;
				// FIXME: Check if this conversion is done above
				if(stride < 0){
					stride *= -1;
					size_t start = lsel->selection[i].u.sub.start;
					lsel->selection[i].u.sub.start = lsel->selection[i].u.sub.finish;
					lsel->selection[i].u.sub.finish = start;
					lsel->selection[i].u.sub.stride = stride;
				}
				size_t gsel_start = gsel->selection[i+1].u.sub.start;
				size_t gsel_finish = gsel->selection[i+1].u.sub.finish;
				size_t lstart = lijk_dims[j].first;
				size_t lfinish = lijk_dims[j].second;
				if((gsel_finish < lstart) || (gsel_start > lfinish)){
					/* The selectors were not designed to represent
					 * 0 elems. This is a hack to represent 0 elems
					 */
					lsel->selection[i].u.sub.start = 0;
					lsel->selection[i].u.sub.finish = -1;
					ldim_sz[i] = 0;	
				}
				else{ 
					if(gsel_start < lstart){
						size_t n_strides = (lstart - gsel_start)/stride + 1;
						lsel->selection[i].u.sub.start = gsel_start + n_strides * stride -lstart;
					}
					if(gsel_finish > lfinish){
						size_t n_strides = (lfinish - lsel->selection[i].u.sub.start)/stride;
						lsel->selection[i].u.sub.finish = lsel->selection[i].u.sub.start + n_strides * stride - lstart;
					}
					if((lsel->selection[i].u.sub.start == lstart) && (lsel->selection[i].u.sub.finish == lfinish)){
						lsel->selection[i].sel_type = Ncl_SUB_ALL;
					}
				}
				/*
				if(gsel->selection[i].u.sub.start >= lijk_dims[j].first){
				}
				if((gsel->selection[i].u.sub.start >= lijk_dims[j].first) && (gsel->selection[i].u.sub.finish <= lijk_dims[j].second)){
					size_t n_strides;
					n_strides = (size_t )ceil((lijk_dims[j].first - gsel->selection[i].u.sub.start)/(gsel->selection[i].u.sub.stride));
					lsel->selection[i].u.sub.start = gsel->selection[i].u.sub.start + n_strides * gsel->selection[i].u.sub.stride;
					n_strides = (size_t )floor((lijk_dims[j].second - lsel->selection[i].u.sub.start)/(lsel->selection[i].u.sub.stride));
					lsel->selection[i].u.sub.finish = lsel->selection[i].u.sub.start + n_strides * lsel->selection[i].u.sub.stride;
					ldim_sz[i] = lijk_dims[j].second - lijk_dims[j].first + 1;
					if((lsel->selection[i].u.sub.start == lijk_dims[j].first) && (lsel->selection[i].u.sub.finish == lijk_dims[j].second)){
						lsel->selection[i].sel_type = Ncl_SUB_ALL;
					}
				}
				else{
					lsel->selection[i].u.sub.start = 0;
					lsel->selection[i].u.sub.finish = 0;
				}
				*/
			}
			else if(sel_type == Ncl_VECSUBSCR){
				size_t n_ind = lsel->selection[i].u.vec.n_ind;
				lsel->selection[i].u.vec.n_ind = 0;	
				for(size_t k=0; k<n_ind; k++){
					long idx = gsel->selection[i+1].u.vec.ind[k];
					if((idx >= lijk_dims[j].first) && (idx <= lijk_dims[j].second)){
						lsel->selection[i].u.vec.ind[lsel->selection[i].u.vec.n_ind] = idx;
						lsel->selection[i].u.vec.n_ind++;
					}
				}
				//ldim_sz[i] = lsel->selection[i].u.vec.n_ind;
			}
			else{
				NhlPError(NhlFATAL, NhlEUNKNOWN, "Unknown selection type, cannot convert to local");
				ldim_sz[i] = 0;
			}
		}
		// Flip the level for local selectors
		if(n_dims >= 3){
			int last_idx = n_dims - 1;
			int lev_idx = last_idx - 2;
			int time_idx = last_idx - 3;
			NclSelection lev_sel = lsel->selection[lev_idx];
			size_t lev_dim_sz = ldim_sz[lev_idx];
			
			for(int k=lev_idx; k<last_idx; k++){
				lsel->selection[k] = lsel->selection[k+1];
				ldim_sz[k] = ldim_sz[k+1];
			}

			lsel->selection[last_idx] = lev_sel;
			ldim_sz[last_idx] = lev_dim_sz;

			if(time_idx >= 0){
				// There are no time steps locally
				ldim_sz[time_idx] = 0;
			}
		}
	}
	else{
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Convert to local failed");
	}	

}

void ScdMBMesh::conform_mesh(NclSelectionRecord *gsel, const std::string &pcal_val_loc)
{
	// If pargal supported sparse tags gsel would just be select ALL
	// Since that is not the case we need to retain the lat/lon selectors
	for(int idx=0; idx < (gsel->n_entries - 1) - 1; idx++){
		if(gsel->selection[idx].sel_type == Ncl_VECSUBSCR){
			NclFree(gsel->selection[idx].u.vec.ind);
		}
		gsel->selection[idx].sel_type = Ncl_SUB_ALL;
	}
}

void ScdMBMesh::conform_mesh(std::vector<size_t> &gdims_sz, const std::string &pcal_val_loc)
{
	size_t n_dims = gdims_sz.size();
	std::vector<std::pair<size_t, size_t> > gij_conform_dims = this->get_gij_conform_dims(pcal_val_loc);
	if((n_dims >= 2) && (gij_conform_dims.size() >= 2)){
		// lon
		gdims_sz[n_dims - 1] = gij_conform_dims[0].second - gij_conform_dims[0].first + 1;
		// lat
		gdims_sz[n_dims - 2] = gij_conform_dims[1].second - gij_conform_dims[1].first + 1;
	}
}

void ScdMBMesh::conform_mesh(std::vector<pargal::pcdim> &gdims, const std::string &pcal_val_loc)
{
	size_t n_dims = gdims.size();
	std::vector<std::pair<size_t, size_t> > gij_conform_dims = this->get_gij_conform_dims(pcal_val_loc);
	if((n_dims >= 2) && (gij_conform_dims.size() >= 2)){
		// lon
		size_t nlon = gij_conform_dims[0].second - gij_conform_dims[0].first + 1;
		gdims[n_dims - 1].set_size(nlon);
		// lat
		size_t nlat = gij_conform_dims[1].second - gij_conform_dims[1].first + 1;
		gdims[n_dims - 2].set_size(nlat);
	}
}

void ScdMBMesh::dbg_print_scd_box(moab::ScdBox *scd_box) const
{
}
} // namespace parncl
